@@ -20,7 +20,9 @@ The recommended way to declare the cache object isto declare it as a global vari
 
 import numpy as np
 import psutil
+
 from copy import deepcopy
+from collections import OrderedDict
 
 def upload_simple_tokens(labels, cache, tensors, grid_setting = False):
     for idx, label in enumerate(labels):
@@ -29,10 +31,7 @@ def upload_simple_tokens(labels, cache, tensors, grid_setting = False):
         else:
             label_completed = (label, (1.0,))
         cache.add(label_completed, tensors[idx])
-#        if type(tensors_scaled) != type(None):
-#            print('uploading scaled data into cache, label: ', label_completed)
-#            cache.add(label_completed, tensors_scaled[idx], scaled = True)        
-        cache.add(label_completed, tensors[idx])        
+        # cache.add(label_completed, tensors[idx])        
         cache.add_base_matrix(label_completed)
 
 
@@ -49,37 +48,46 @@ def upload_grids(grids, cache):
     upload_simple_tokens(labels = labels, cache = cache, tensors = tensors, grid_setting = True)
     cache.use_structural()
 
-def np_ndarray_section(matrix, boundary, except_idx : list = []):
-    if boundary != 0:
+
+def np_ndarray_section(matrix, boundary = None, except_idx : list = []):
+    #Добавить проверку на случай, когда boundary = 0, чтобы тогда возвращало матрицу без изменения
+    if isinstance(boundary, int):
+        if boundary != 0:
+            for idx in except_idx:
+                if idx < 0:
+                    except_idx.append(matrix.ndim - 1)
+            for dim_idx in np.arange(matrix.ndim):
+                if not dim_idx in except_idx:
+                    matrix = np.moveaxis(matrix, source = dim_idx, destination=0)
+                    matrix = matrix[boundary:-boundary, ...]
+                    matrix = np.moveaxis(matrix, source = 0, destination=dim_idx)
+    elif isinstance(boundary, (list, tuple)):
+        if len(boundary) != matrix.ndim:
+            raise IndexError('Sizes of boundary do not match the dimensionality of data')
         for idx in except_idx:
             if idx < 0:
                 except_idx.append(matrix.ndim - 1)
         for dim_idx in np.arange(matrix.ndim):
-            if not dim_idx in except_idx:
+            if not dim_idx in except_idx and boundary[dim_idx] > 0:
                 matrix = np.moveaxis(matrix, source = dim_idx, destination=0)
-                matrix = matrix[boundary:-boundary, ...]
-                matrix = np.moveaxis(matrix, source = 0, destination=dim_idx)
+                matrix = matrix[boundary[dim_idx]:-boundary[dim_idx], ...]
+                matrix = np.moveaxis(matrix, source = 0, destination=dim_idx)    
     return matrix
-    
 
-def prepare_var_tensor(var_tensor, derivs_tensor, time_axis, boundary, cut_except = [], axes = None,
-                       cut_axes : bool = True):
+    
+def prepare_var_tensor(var_tensor, derivs_tensor, time_axis, boundary, cut_except = []):
     initial_shape = var_tensor.shape
+    print('initial_shape', initial_shape, 'derivs_tensor.shape', derivs_tensor.shape)
     var_tensor = np.moveaxis(var_tensor, time_axis, 0)
-    if axes is not None:
-        if all(ax.shape == var_tensor.shape for ax in axes):
-            axes = [np_ndarray_section(ax, boundary, cut_except) if cut_axes else ax for ax in axes]
-        elif not all(ax.shape == (shape - 2*boundary for shape in var_tensor.shape) for ax in axes):
-#            print(ax.shape, (shape - 2*boundary for shape in var_tensor.shape))
-            raise ValueError('Shapes of axes not fitting')
-        increment = len(axes) + 1
-        result = np.ones((increment + derivs_tensor.shape[-1], ) + tuple([shape - 2*boundary for shape in var_tensor.shape]))
-        for ax_idx, ax in enumerate(axes):
-            result[ax_idx, ... ] = ax
-#    print(result.shape, var_tensor.shape, boundary, initial_shape)
-    else:
+    if isinstance(boundary, int):
         result = np.ones((1 + derivs_tensor.shape[-1], ) + tuple([shape - 2*boundary for shape in var_tensor.shape]))
-        increment = 1
+    elif isinstance(boundary, (tuple, list)):
+        result = np.ones((1 + derivs_tensor.shape[-1], ) + tuple([shape - 2*boundary[dim_idx] for dim_idx, 
+                                                                  shape in enumerate(var_tensor.shape)]))
+    else:
+        raise TypeError('')
+    increment = 1
+    print('var_tensor.shape', var_tensor.shape, np_ndarray_section(var_tensor, boundary, cut_except).shape, boundary)
     result[increment - 1, :] = np_ndarray_section(var_tensor, boundary, cut_except)
     if derivs_tensor.ndim == 2:
         for i_outer in range(0, derivs_tensor.shape[1]):
@@ -94,8 +102,7 @@ def prepare_var_tensor(var_tensor, derivs_tensor, time_axis, boundary, cut_excep
     return result
 
 
-def download_variable(var_filename, deriv_filename, boundary, time_axis = 0):
-#    print('loading', var_filename, deriv_filename)
+def download_variable(var_filename, deriv_filename, boundary, time_axis):
     var = np.load(var_filename)
     derivs = np.load(deriv_filename)
     tokens_tensor = prepare_var_tensor(var, derivs, time_axis, boundary)
@@ -104,17 +111,12 @@ def download_variable(var_filename, deriv_filename, boundary, time_axis = 0):
 
 class Cache(object):
     def __init__(self):
+        self.max_allowed_tensors = None
         self.memory_default = dict()
         self.memory_normalized = dict()
         self.memory_structural = dict()
-        # if use_structural else False
         self.mem_prop_set = False
         self.base_tensors = [] #storage of non-normalized tensors, that will not be affected by change of variables
-#        self.base_tensors_scaled = dict()
-#        
-#    def merge_structural_and_base(self, except_labels = []):
-#        for key in self.memory_structural.keys():
-#            self.structural_and_base_merged[key] = True if not key in except_labels else True
 
     def use_structural(self, use_base_data = True, label = None, replacing_data = None):
         assert use_base_data or not type(replacing_data) == type(None), 'Structural data must be declared with base data or by additional tensors.'
@@ -236,8 +238,10 @@ class Cache(object):
 #        assert not scaled or self.scale_used, 'Trying to add scaled data, while the cache it not allowed to get it'
 #        assert not structural, 'the structural data must be added with cache.use_structural method'
         if normalized:
+            if self.max_allowed_tensors is None:
+                self.memory_usage_properties(obj_test_case = tensor, mem_for_cache_frac = 5)
             if (len(self.memory_normalized) + len(self.memory_default) + len(self.memory_structural) < self.max_allowed_tensors and 
-                label not in self.memory_default.keys()):
+                label not in self.memory_normalized.keys()):
                 self.memory_normalized[label] = tensor
                 if indication: 
                     print('Enough space for saved normalized term ', label, tensor.nbytes)
@@ -249,11 +253,13 @@ class Cache(object):
                 return True
             else:
                 if indication: 
-                    print('Not enough space for term ', label, tensor.nbytes)
+                    print('Not enough space for term ', label, tensor.nbytes, 'Can save only', self.max_allowed_tensors, 'tensors. While already uploaded ', len(self.memory_normalized) + len(self.memory_default) + len(self.memory_structural))
                 return False         
         elif structural:
             raise NotImplementedError('The structural data must be added with cache.use_structural method')
         else:
+            if self.max_allowed_tensors is None:
+                self.memory_usage_properties(obj_test_case = tensor, mem_for_cache_frac = 5)            
             if (len(self.memory_normalized) + len(self.memory_default) + len(self.memory_structural) < self.max_allowed_tensors and 
                 label not in self.memory_default.keys()):
                 self.memory_default[label] = tensor
@@ -311,6 +317,21 @@ class Cache(object):
                 print('fetched label:', label, ' prev. known as ', saved_as)
                 raise KeyError('Can not fetch tensor from cache with non-normalied data')
 
+    def get_all(self, normalized = False, structural = False):
+        if normalized:
+            processed_mem = self.memory_normalized
+        elif structural:
+            processed_mem = self.memory_structural
+        else:
+            processed_mem = self.memory_default
+
+        keys = []; tensors = []
+        for key, value in processed_mem.items():
+            keys.append(key)
+            tensors.append(value)
+        
+        return keys, tensors
+            
     def __contains__(self, obj):
         '''
         Valid input type:
@@ -318,14 +339,13 @@ class Cache(object):
             np.ndarray of values (checked in unnormalized data); (np.ndarray, normalized), where normalized is bool 
             (T if norm, else F) and np.ndarray is np.ndarray of tensor values. Does not support scaled vals
         '''
-#        raise NotImplementedError
-        if type(obj) == str:
+        if (type(obj) == tuple or type(obj) == list) and type(obj[0]) == str:
             return obj in self.memory_default.keys()
-        elif (type(obj) == tuple or type(obj) == list) and type(obj[0]) == str and type(obj[1]) == bool:
+        elif (type(obj) == tuple or type(obj) == list) and type(obj[0]) == tuple and type(obj[1]) == bool:
             if obj[1]:
-                return obj in self.memory_normalized.keys()
+                return obj[0] in self.memory_normalized.keys()
             else:
-                return obj in self.memory_default.keys()
+                return obj[0] in self.memory_default.keys()
         elif type(obj) == np.ndarray:
             try:
                 return np.any([np.all(obj == entry_values) for entry_values in self.memory_default.values()])
@@ -341,7 +361,21 @@ class Cache(object):
                 return False
         else:
             raise NotImplementedError('Invalid format of function input to check, if the object is in cache')
-            
+    
+#    def __iter__(self):
+#        for key in self.memory_default.keys()
+    def prune_tensors(self, pruner, mem_to_process : list = ['default', 'structural', 'normalized']):
+        mem_arranged = {'default' : self.memory_default, 
+                            'structural' : self.memory_structural, 
+                            'normalized' : self.memory_normalized}
+        
+        for key in self.memory_default.keys():
+            for mem_type in mem_to_process:
+                try:
+                    mem_arranged[mem_type][key] = pruner.prune(mem_arranged[mem_type][key])
+                except (NameError, KeyError) as e:
+                    pass
+        
     @property
     def consumed_memory(self):
         memsize = np.sum([value.nbytes for _, value in self.memory_default.items()])
@@ -349,3 +383,21 @@ class Cache(object):
         for label, merged_state in self.structural_and_base_merged.items():
             if not merged_state: memsize += self.memory_structural[label].nbytes
         return memsize
+    
+
+def upload_complex_token(label : str, params_values : OrderedDict, evaluator, tensor_cache : Cache, grid_cache : Cache):
+    # label_completed = (label, (1.0,))
+    # value = evaluator.apply(self)
+    # global_var.tensor_cache.add(self.cache_label, value, structural = False)
+    try:
+        evaluation_function = evaluator.evaluation_functions[label] 
+    except TypeError:
+        evaluation_function = evaluator.evaluation_functions    
+    _, grids = grid_cache.get_all()
+    grid_function = np.vectorize(lambda args: evaluation_function(*args, **params_values))
+    indexes_vect = np.empty_like(grids[0], dtype = object)
+    for tensor_idx, _ in np.ndenumerate(grids[0]):
+        indexes_vect[tensor_idx] = tuple([grid[tensor_idx] for grid in grids])    
+        
+    label_completed = (label, tuple(params_values.values()))
+    tensor_cache.add(label_completed, grid_function(indexes_vect))
\ No newline at end of file
