@@ -10,6 +10,10 @@ import numpy as np
 from functools import reduce
 import copy
 import gc
+import time
+import datetime
+import pickle
+import warnings
 
 import epde.globals as global_var
 
@@ -31,10 +35,11 @@ def normalize_ts(Input):
 #    print(Matrix.shape)
     if np.ndim(matrix) == 0:
         raise ValueError('Incorrect input to the normalizaton: the data has 0 dimensions')
-    elif np.ndim(matrix) == 0:
+    elif np.ndim(matrix) == 1:
         return matrix
     else:
         for i in np.arange(matrix.shape[0]):
+            # print(matrix[i].shape)
             std = np.std(matrix[i])
             if std != 0:
                 matrix[i] = (matrix[i] - np.mean(matrix[i])) / std
@@ -80,24 +85,24 @@ class Complex_Structure(object):
 
 
 class Term(Complex_Structure):
-    def __init__(self, pool, 
-                 passed_term = None, max_factors_in_term = 1, forbidden_tokens = None, interelement_operator = np.multiply):
+    __slots__ = ['_history', 'structure', 'interelement_operator', 'saved', 'saved_as', 
+                 'pool', 'max_factors_in_term', 'cache_linked', 'occupied_tokens_labels']
+    
+    def __init__(self, pool, passed_term = None, max_factors_in_term = 1, forbidden_tokens = None, 
+                 interelement_operator = np.multiply):
         super().__init__(interelement_operator)
         self.pool = pool
         self.max_factors_in_term = max_factors_in_term
         
-        if type(passed_term) == type(None):
+        if passed_term is None:
             self.Randomize(forbidden_tokens)     
         else:
             self.Defined(passed_term)
 
-        if type(global_var.tensor_cache) != type(None):
+        if global_var.tensor_cache is not None:
             self.use_cache()
         self.reset_saved_state() # key - state of normalization, value - if the variable is saved in cache
-        
-#    def reset_saved_state(self):
-#        super().reset_saved_state()
-#        self.saved_solver_form = False
+
         
     @property
     def cache_label(self):
@@ -116,24 +121,24 @@ class Term(Complex_Structure):
                 self.structure[idx].use_cache()
 
 
-    def Defined(self, passed_term):   
+    def Defined(self, passed_term):
         self.structure = []
         print('passed_term:', passed_term)
 
-        if type(passed_term) == list or type(passed_term) == tuple:
+        if isinstance(passed_term, (list, tuple)): #type(passed_term) == list or type(passed_term) == tuple:
             for i, factor in enumerate(passed_term):
-                if type(factor) == str:
+                if isinstance(factor, str):
                     _, temp_f = self.pool.create(label = factor)
                     self.structure.append(temp_f)#; raise NotImplementedError
-                elif type(factor) == Factor:
+                elif isinstance(factor, Factor):
                     self.structure.append(factor)
                 else:
                     raise ValueError('The structure of a term should be declared with str or factor.Factor obj, instead got', type(factor))
         else:   # Случай, если подается лишь 1 токен
-            if type(passed_term) == str:
+            if isinstance(passed_term, str):
                 _, temp_f = self.pool.create(label = passed_term)
                 self.structure.append(temp_f)#; raise NotImplementedError
-            elif type(passed_term) == Factor:
+            elif isinstance(passed_term, Factor):
                 self.structure.append(passed_term)
             else:
                 raise ValueError('The structure of a term should be declared with str or factor.Factor obj, instead got', type(passed_term))
@@ -143,32 +148,42 @@ class Term(Complex_Structure):
         if np.sum(self.pool.families_cardinality(meaningful_only = True)) == 0:
             raise ValueError('No token families are declared as meaningful for the process of the system search')
         factors_num = np.random.randint(1, self.max_factors_in_term +1)
-#        print('factors:', factors_num)
-        while True:        
+        while True:
             self.occupied_tokens_labels = []
             occupied_by_factor, factor = self.pool.create(label = None, create_meaningful = True, 
                                                            occupied = self.occupied_tokens_labels, **kwargs)
             self.structure = [factor,]
             self.occupied_tokens_labels.extend(occupied_by_factor)
+            factors_powers = {factor.label : 1}
             
             for i in np.arange(1, factors_num):
                 occupied_by_factor, factor = self.pool.create(label = None, create_meaningful = False, 
                                                                occupied = self.occupied_tokens_labels, 
                                                                def_term_tokens = [token.label for token in self.structure], 
                                                                **kwargs) 
+                if factor.label in factors_powers:
+                    factors_powers[factor.label] += 1
+                else:
+                    factors_powers[factor.label] = 1
+                    
+                for param_idx, param_descr in factor.params_description.items():
+                    if param_descr['name'] == 'power': power_param_idx = param_idx
+                        
+                if factors_powers[factor.label] == factor.params_description[power_param_idx]['bounds'][1]:
+                    self.occupied_tokens_labels.append(factor.label)
                 self.structure.append(factor)
                 self.occupied_tokens_labels.extend(occupied_by_factor)                
             self.structure = Filter_powers(self.structure)
-            if type(forbidden_factors) == type(None):
+            if forbidden_factors is None:
                 break
             elif all([(Check_Unqueness(factor, forbidden_factors) or not factor.status['unique_for_right_part']) for factor in self.structure]):
                 break            
 
 
     def evaluate(self, structural):
-        assert type(global_var.tensor_cache) != type(None), 'Currently working only with connected cache'
+        assert global_var.tensor_cache is not None, 'Currently working only with connected cache'
         normalize = structural
-        if self.saved[structural]:
+        if self.saved[structural] or (self.cache_label, normalize) in global_var.tensor_cache: #  
             value = global_var.tensor_cache.get(self.cache_label, normalized = normalize,
                                                 saved_as = self.saved_as[normalize]) 
             value = value.reshape(value.size)
@@ -184,10 +199,40 @@ class Term(Complex_Structure):
                 value = (value - np.mean(value))   
             if np.all([len(factor.params) == 1 for factor in self.structure]):
                 self.saved[normalize] = global_var.tensor_cache.add(self.cache_label, value, normalized = normalize) # Место возможных проблем: сохранение/загрузка нормализованных данных
-                if self.saved[normalize]: self.saved_as[normalize] = self.cache_label 
+                if self.saved[normalize]: self.saved_as[normalize] = self.cache_label
             value = value.reshape(value.size)
             return value            
 
+    def Filter_tokens_by_right_part(self, reference_target, equation, equation_position):
+        taken_tokens = [factor.label for factor in reference_target.structure if factor.status['unique_for_right_part']]
+        meaningful_taken = any([factor.status['meaningful'] for factor in reference_target.structure
+                            if factor.status['unique_for_right_part']])
+        
+        accept_term_try = 0
+        while True:
+            accept_term_try += 1
+            new_term = copy.deepcopy(self)            
+            for factor_idx, factor in enumerate(new_term.structure):
+                if factor.label in taken_tokens:
+                    new_term.Reset_occupied_tokens()
+                    _, new_term.structure[factor_idx] = self.pool.create(create_meaningful=meaningful_taken, 
+                                                             occupied = new_term.occupied_tokens_labels + taken_tokens)
+                    # print('try:', accept_term_try, 'suggested:', new_term.name)
+                    #Возможная ошибка из-за (не) использования "create meaningful"
+            if Check_Unqueness(new_term, equation.structure[:equation_position] + 
+                                         equation.structure[equation_position + 1 :]):
+                self.structure = new_term.structure
+                self.structure = Filter_powers(self.structure)
+                self.reset_saved_state()
+                break
+            if accept_term_try == 10 and global_var.verbose.show_warnings:
+                warnings.warn('Can not create unique term, while filtering equation tokens in regards to the right part.')
+            if accept_term_try >= 10:
+                self.Randomize(forbidden_factors = new_term.occupied_tokens_labels + taken_tokens)
+            if accept_term_try == 100:
+                print('Something wrong with the random generation of term while running "Filter_tokens_by_right_part"')
+                print('proposed', new_term.name, 'for ', equation.text_form, 'with respect to', reference_target.name)
+
     def Reset_occupied_tokens(self):
         occupied_tokens_new = []
         for factor in self.structure:
@@ -196,11 +241,6 @@ class Term(Complex_Structure):
                     occupied_tokens_new.extend([token for token in token_family.tokens])
                 elif factor.status['unique_specific_token']:
                     occupied_tokens_new.append(factor.label)
-#            for token_family in self.tokens:
-#                if factor.label in token_family.tokens and token_family.status['unique_token_type']:
-#                    occupied_tokens_new.extend([token for token in token_family.tokens])
-#                elif factor.label in token_family.tokens and token_family.status['unique_specific_token']:
-#                    occupied_tokens_new.append(factor.label)
         self.occupied_tokens_labels = occupied_tokens_new
 
     @property
@@ -226,6 +266,10 @@ class Term(Complex_Structure):
                 form += ' * '
         return form
     
+    @property
+    def contains_deriv(self):
+        return any([factor.is_deriv and factor.deriv_code != [None,] for factor in self.structure])
+    
     @property
     def solver_form(self):
         
@@ -247,21 +291,48 @@ class Term(Complex_Structure):
             deriv_orders = deriv_orders[0]
             
         coeff_tensor = torch.from_numpy(coeff_tensor)
-        return [coeff_tensor, deriv_powers, deriv_orders]
-                
+        return [coeff_tensor, deriv_orders, deriv_powers]
                 
     def __eq__(self, other):
-#        if type(other) != type(self):
-#            print(type(self), type(other), isinstance(self, type(other)))
-#            raise TypeError('Equality check between terms failed. Wrong type passed')
         return (all([any([other_elem == self_elem for other_elem in other.structure]) for self_elem in self.structure]) and 
                 all([any([other_elem == self_elem for self_elem in self.structure]) for other_elem in other.structure]) and 
                 len(other.structure) == len(self.structure))
 
+    @History_Extender('\n -> was copied by deepcopy(self)', 'n')
+    def __deepcopy__(self, memo = None):
+        clss = self.__class__
+        new_struct = clss.__new__(clss)
+        memo[id(self)] = new_struct
+        
+        attrs_to_avoid_copy = [] # Add attributes to avoid copying during deepcopy, if needed
+        for k in self.__slots__:
+            try:
+                if k not in attrs_to_avoid_copy:
+                    if not isinstance(k, list):
+                        setattr(new_struct, k, copy.deepcopy(getattr(self, k), memo))
+                    else:
+                        temp = []
+                        for elem in getattr(self, k):
+                            temp.append(copy.deepcopy(elem, memo))
+                        setattr(new_struct, k, temp)
+                else:
+                    setattr(new_struct, k, None)
+            except AttributeError:
+                pass
 
-class Equation(Complex_Structure):
-#    @Reset_equation_status()   
-    def __init__(self, pool, basic_structure, terms_number = 6, max_factors_in_term = 2, interelement_operator = np.add): #eq_weights_eval
+        return new_struct
+
+
+class Equation(Complex_Structure): 
+    __slots__ = ['_history', 'structure', 'interelement_operator', 'saved', 'saved_as', 
+                 'n_immutable', 'pool', 'terms_number', 'max_factors_in_term', 'operator', 
+                  '_target', 'target_idx', '_features', 'right_part_selected', 
+                  '_weights_final', 'weights_final_evald', '_weights_internal', 'weights_internal_evald', 
+                  'fitness_calculated', 'solver_form_defined', '_solver_form', '_fitness_value', 
+                  'crossover_selected_times', 'elite']
+
+    def __init__(self, pool, basic_structure, terms_number = 6, max_factors_in_term = 2, 
+                 interelement_operator = np.add): #eq_weights_eval
 
         """
 
@@ -304,15 +375,21 @@ class Equation(Complex_Structure):
         self.reset_state()
 
         self.n_immutable = len(basic_structure)
+        # print('n_immutable', self.n_immutable)
         self.pool = pool
         self.structure = []
         self.terms_number = terms_number; self.max_factors_in_term = max_factors_in_term
         self.operator = None
         if (terms_number < self.n_immutable): 
             raise Exception('Number of terms ({}) is too low to even contain all of the pre-determined ones'.format(terms_number))        
+
+        for passed_term in basic_structure:
+            if isinstance(passed_term, Term):
+                self.structure.append(passed_term)
+            elif isinstance(passed_term, str):
+                self.structure.append(Term(self.pool, passed_term = passed_term, 
+                                           max_factors_in_term = self.max_factors_in_term))
             
-        self.structure.extend([Term(self.pool, passed_term = label, max_factors_in_term = self.max_factors_in_term) for label in basic_structure])
-    
         for i in range(len(basic_structure), terms_number):
             check_test = 0
             while True:
@@ -323,56 +400,12 @@ class Equation(Complex_Structure):
             self.structure.append(new_term)
 
         for idx, _ in enumerate(self.structure):
-            self.structure[idx].use_cache()    
-
-    def select_target_idx(self, separate_vars = [], operator = None, target_idx_fixed = None): 
-        '''
-        
-        Separation of target term from features & removal of factors, that are in target, from features
-        
-        '''
-        raise NotImplementedError
-#        assert (type(self.operator) != type(None) or type(operator) != type(None)) or type(target_idx_fixed) != type(None)
-#        if type(self.operator) == type(None) and type(operator) != type(None): self.operator = operator
-#
-#        if type(target_idx_fixed) == type(None):
-#            max_fitness = 0
-#            max_idx = 0
-#            for target_idx, _ in enumerate(self.structure): # target_term
-#                self.target_idx = target_idx
-##                print('target:', self.structure[self.target_idx].name)
-#                self.operator.get_fitness(self)
-#                if self.described_variables in separate_vars:
-##                    print('penalizing')
-#                    self.penalize_fitness(coeff = 0.) 
-##                print('idx:', self.target_idx, self.fitness_value, max_fitness, max_idx, '\n\t', self.weights_internal, self.weights_final)
-##                print(self.text_form)                     
-#                if self.fitness_value > max_fitness: 
-#                    max_fitness = self.fitness_value
-#                    max_idx = target_idx                  # Ошибка, когда target_idx == 0
-#            self.target_idx = max_idx
-#            self.operator.get_fitness(self)
-#            eps = 1e-5
-#            assert np.abs(self.fitness_value - max_fitness) < eps
-#        else:
-#            self.target_idx = target_idx_fixed
-#      
-
-    def check_split_correctness(self): #  Refactor for needs of system discovery
-        pass
-#        if True:
-#            pass
-#        else:
-#            target = self.structure[self.target_idx]
-#            for term_idx in range(len(self.structure)):
-#                if term_idx == self.target_idx: continue
-#                for factor in self.structure[term_idx].structure:
-#                    if factor.token.status['unique_for_right_part'] and any([factor == rp_factor for rp_factor in target.structure]):
-#                        print('In target:', target.name) 
-#                        print('In other structure discovered:', [factor.label for factor in self.structure[term_idx].structure])
-#                        print([term.name for term in self.structure])
-#                        raise ValueError('Forbidden factor ')                            
+            self.structure[idx].use_cache()            
         
+    @property
+    def contains_deriv(self):
+        return any([term.contains_deriv for term in self.structure])
+    
     @property 
     def forbidden_token_labels(self):
         target_symbolic = [factor.label for factor in self.structure[self.target_idx].structure]
@@ -380,10 +413,30 @@ class Equation(Complex_Structure):
 
         for token_family in self.pool.families:
             for token in token_family.tokens:
-                if token in target_symbolic and token.status['unique_for_right_part']:
+                if token in target_symbolic and token_family.status['unique_for_right_part']:
                     forbidden_tokens.add(token)        
         return forbidden_tokens
 
+    def reconstruct_to_contain_deriv(self):
+        while True:
+            replacement_idx = np.random.randint(low = 0, high = len(self.structure))
+            temp = Term(self.pool, max_factors_in_term=self.max_factors_in_term) # ,  forbidden_tokens=self.forbidden_token_labels
+            if temp.contains_deriv:
+                self.structure[replacement_idx] = temp
+                break
+
+    def reconstruct_by_right_part(self, right_part_idx):
+        new_eq = copy.deepcopy(self)
+        self.copy_properties_to(new_eq)
+        new_eq.target_idx = right_part_idx
+        if any([factor.status['unique_for_right_part'] for factor in new_eq.structure[right_part_idx].structure]):
+            for term_idx, term in enumerate(new_eq.structure):
+                if term_idx != right_part_idx:
+                    term.Filter_tokens_by_right_part(new_eq.structure[right_part_idx], self, term_idx)
+
+        new_eq.reset_saved_state()
+        return new_eq
+
     def evaluate(self, normalize = True, return_val = False, save = True):
         self._target = self.structure[self.target_idx].evaluate(normalize)
         feature_indexes = list(range(len(self.structure)))
@@ -396,6 +449,8 @@ class Equation(Complex_Structure):
                 self._features = np.vstack([self._features, temp])
             else:
                 continue
+        if self._features.ndim == 1:
+            self._features = np.expand_dims(self._features, 1)
         temp_feats = np.vstack([self._features, np.ones(self._features.shape[1])])
         self._features = np.transpose(self._features); temp_feats = np.transpose(temp_feats)
         if return_val:
@@ -412,7 +467,7 @@ class Equation(Complex_Structure):
             return value, self._target, self._features
         else:
             return None, self._target, self._features
-            
+        
     def reset_state(self, reset_right_part : bool = True):
         if reset_right_part: self.right_part_selected = False
         self.weights_internal_evald = False
@@ -421,14 +476,41 @@ class Equation(Complex_Structure):
         self.solver_form_defined = False
     
     @Reset_equation_status(reset_input = False, reset_output = True)
-    @History_Extender(f'\n -> was copied by deepcopy(self)', 'n')
+    @History_Extender('\n -> was copied by deepcopy(self)', 'n')
     def __deepcopy__(self, memo = None):
         clss = self.__class__
         new_struct = clss.__new__(clss)
-        new_struct.__dict__.update(self.__dict__)        
-#        print('new_struct:', type(new_struct), new_struct)
+        memo[id(self)] = new_struct
+        
+        attrs_to_avoid_copy = ['_features', '_target']
+        for k in self.__slots__:
+            try:
+                if k not in attrs_to_avoid_copy:
+                    if not isinstance(k, list):
+                        setattr(new_struct, k, copy.deepcopy(getattr(self, k), memo))
+                    else:
+                        temp = []
+                        for elem in getattr(self, k):
+                            temp.append(copy.deepcopy(elem, memo))
+                        setattr(new_struct, k, temp)
+                else:
+                    setattr(new_struct, k, None)
+            except AttributeError:
+                pass
+
         return new_struct
             
+    def copy_properties_to(self, new_equation):
+        new_equation.weights_internal_evald = self.weights_internal_evald
+        new_equation.weights_final_evald = self.weights_final_evald
+        new_equation.right_part_selected = self.right_part_selected
+        new_equation.fitness_calculated = self.fitness_calculated
+        new_equation.solver_form_defined = self.solver_form_defined
+        try:
+            new_equation._fitness_value = self._fitness_value
+        except AttributeError:
+            pass
+            
     def add_history(self, add):
         self._history += add
             
@@ -446,10 +528,6 @@ class Equation(Complex_Structure):
     
     def penalize_fitness(self, coeff = 1.):
         self._fitness_value = self._fitness_value*coeff
-#    
-#    @property
-#    def L0_norm(self):
-#        return np.count_nonzero(self.weights_internal)
 
     @property
     def weights_internal(self):    
@@ -459,20 +537,21 @@ class Equation(Complex_Structure):
             raise AttributeError('Internal weights called before initialization')
         
     @weights_internal.setter
-    def weights_internal(self, weights):    # Ошибка!
+    def weights_internal(self, weights):
         self._weights_internal = weights
         self.weights_internal_evald = True
         self.weights_final_evald = False
         
     @property
-    def weights_final(self): # ошибка из-за одинаковых фин. весов для различных индексов правых частей уравнений
-        if self.weights_final_evald: # Настроить переоценку после любых изменений структуры уравнения
+    def weights_final(self): 
+        if self.weights_final_evald:
             return self._weights_final
         else:
+            print(self.text_form)
             raise AttributeError('Final weights called before initialization')
             
     @weights_final.setter
-    def weights_final(self, weights):    # Ошибка!
+    def weights_final(self, weights):
         self._weights_final = weights
         self.weights_final_evald = True
             
@@ -489,11 +568,16 @@ class Equation(Complex_Structure):
     @property
     def text_form(self):
         form = ''
-        for term_idx in range(len(self.structure)):
-            if term_idx != self.target_idx:
-                form += str(self.weights_final[term_idx]) if term_idx < self.target_idx else str(self.weights_final[term_idx-1])
-                form += ' * ' + self.structure[term_idx].name + ' + '
-        form += str(self.weights_final[-1]) + ' = ' + self.structure[self.target_idx].name
+        if self.weights_final_evald:
+            for term_idx in range(len(self.structure)):
+                if term_idx != self.target_idx:
+                    form += str(self.weights_final[term_idx]) if term_idx < self.target_idx else str(self.weights_final[term_idx-1])
+                    form += ' * ' + self.structure[term_idx].name + ' + '
+            form += str(self.weights_final[-1]) + ' = ' + self.structure[self.target_idx].name            
+        else:
+            for term_idx in range(len(self.structure)):
+                    form += 'k_'+ str(term_idx) + ' ' + self.structure[term_idx].name + ' + '
+            form +=  'k_' + str(len(self.structure)) + ' = 0'
         return form
 
     def solver_form(self):
@@ -506,16 +590,19 @@ class Equation(Complex_Structure):
                     term_form = self.structure[term_idx].solver_form
                     weight = self.weights_final[term_idx] if term_idx < self.target_idx else self.weights_final[term_idx-1]
                     term_form[0] = term_form[0] * weight
+                    term_form[0] = torch.flatten(term_form[0]).unsqueeze(1).type(torch.FloatTensor)
                     self._solver_form.append(term_form)
                     
             free_coeff_weight = torch.from_numpy(np.full_like(a = global_var.grid_cache.get('0'), 
                                                               fill_value = self.weights_final[-1]))
+            free_coeff_weight = torch.flatten(free_coeff_weight).unsqueeze(1).type(torch.FloatTensor)
             target_weight = torch.from_numpy(np.full_like(a = global_var.grid_cache.get('0'), 
                                                               fill_value = -1))            
             target_form = self.structure[self.target_idx].solver_form
             target_form[0] = target_form[0] * target_weight
+            target_form[0] = torch.flatten(target_form[0]).unsqueeze(1).type(torch.FloatTensor)
             
-            self._solver_form.append([free_coeff_weight, None, 0])
+            self._solver_form.append([free_coeff_weight, [None,], 0])
             self._solver_form.append(target_form)
             self.solver_form_defined = True            
             return self._solver_form
@@ -534,87 +621,103 @@ class Equation(Complex_Structure):
             else:
                 weight_idx = term_idx if term_idx < term_idx else term_idx - 1
                 if np.abs(self.weights_final[weight_idx]) > eps:
-#                    print('to described', self.weights_final[weight_idx], {factor.token.type for factor in term.structure})
                     described.update({factor.type for factor in term.structure})
         described = frozenset(described)
         return described
     
-    def required_BC(self):
-        solver_form = self.solver_form
+    def max_deriv_orders(self):
+        solver_form = self.solver_form()
         max_orders = np.zeros(global_var.grid_cache.get('0').ndim)
+        
+        def count_order(obj, deriv_ax):
+            if obj is None:
+                return 0
+            else:
+                return obj.count(deriv_ax)
+                
         for term in solver_form:
             if isinstance(term[2], list):
                 for deriv_factor in term[1]:
-                    orders = np.array([deriv_factor.count(ax) for ax 
-                                       in np.arange(max_orders.shape)])
+                    orders = np.array([count_order(deriv_factor, ax) for ax #deriv_factor.count(ax)
+                                       in np.arange(max_orders.size)])
                     max_orders = np.maximum(max_orders, orders)
             else:
-                orders = np.array([term[1].count(ax) for ax 
-                                   in np.arange(max_orders.shape)])
+                orders = np.array([count_order(term[1], ax) for ax # term[1].count(ax)
+                                   in np.arange(max_orders.size)])
                 max_orders = np.maximum(max_orders, orders)
-        if np.max(max_orders) <= 2:
+        if np.max(max_orders) > 2:
             raise NotImplementedError('The current implementation allows does not allow higher orders of equation, than 2.')
         return max_orders
+    
+    def boundary_conditions(self, main_var_key = ('u', (1.0,))):
+        required_bc_ord = self.max_deriv_orders()   # We assume, that the maximum order of the equation here is 2
+        if global_var.grid_cache is None:
+            raise NameError('Grid cache has not been initialized yet.')
+        
+        bconds = []
+        hardcoded_bc_relative_locations = {0 : None, 1 : (0,), 2 : (0, 1)}
+        tensor_shape = global_var.grid_cache.get('0').shape
 
-#def Get_true_coeffs(equation): # Не забыть про то, что последний коэф - для константы
-#    assert equation.weights_internal_evald, 'Trying to calculate final weights before evaluating intermeidate ones (no sparcity).'
-#    target = equation.structure[equation.target_idx]
-#
-#    equation.check_split_correctness()
-#            
-##    print(type(target.value))
-#    target_vals = target.evaluate(False)
-#    features_vals = []
-#    nonzero_features_indexes = []
-#    for i in range(len(equation.structure)):
-#        if i == equation.target_idx:
-#            continue
-#        idx = i if i < equation.target_idx else i-1
-#        if equation.weights_internal[idx] != 0:
-#            features_vals.append(equation.structure[i].evaluate(False))
-#            nonzero_features_indexes.append(idx)
-#            
-##    print('Indexes of nonzero elements:', nonzero_features_indexes)
-#    if len(features_vals) == 0:
-#        return np.zeros(len(equation.structure)) #Bind_Params([(token.label, token.params) for token in target.structure]), [('0', 1)]
-#    
-#    features = features_vals[0]
-#    if len(features_vals) > 1:
-#        for i in range(1, len(features_vals)):
-#            features = np.vstack([features, features_vals[i]])
-#    features = np.vstack([features, np.ones(features_vals[0].shape)]) # Добавляем константную фичу
-#    features = np.transpose(features)  
-#    
-#    estimator = LinearRegression(fit_intercept=False)
-#    try:
-#        estimator.fit(features, target_vals)
-#    except ValueError:
-#        features = features.reshape(-1, 1)
-#        estimator.fit(features, target_vals)
-#        
-#    valueable_weights = estimator.coef_
-#    weights = np.zeros(len(equation.structure))
-#    for weight_idx in range(len(weights)-1):
-#        if weight_idx in nonzero_features_indexes:
-#            weights[weight_idx] = valueable_weights[nonzero_features_indexes.index(weight_idx)]
-#    weights[-1] = valueable_weights[-1]    
-##    print('weights check:', weights, equation.weights_internal)
-#    return weights
+        def get_boundary_ind(tensor_shape, axis, rel_loc):
+            return tuple(np.meshgrid(*[np.arange(shape) if dim_idx != axis else min(int(rel_loc * shape), shape-1)
+                       for dim_idx, shape in enumerate(tensor_shape)], indexing = 'ij'))
+
+        for ax_idx, ax_ord in enumerate(required_bc_ord):
+
+            for loc_fraction in hardcoded_bc_relative_locations[ax_ord]:
+                indexes = get_boundary_ind(tensor_shape, axis = ax_idx, rel_loc = loc_fraction)
+                coords = np.squeeze(np.array([global_var.grid_cache.get(str(idx))[indexes] for idx in np.arange(len(tensor_shape))])).T
+                vals = np.squeeze(global_var.tensor_cache.get(main_var_key)[indexes]).T
+                
+                coords = torch.from_numpy(coords).type(torch.FloatTensor)
+                vals = torch.from_numpy(vals).type(torch.FloatTensor)
+                bconds.append([coords, vals])    
+
+        print('shape of the grid', global_var.grid_cache.get('0').shape)
+        print('Obtained boundary conditions', len(bconds[0]))
+        return bconds
+
+    
+def standalone_boundary_conditions(max_deriv_orders, main_var_key = ('u', (1.0,))):
+    required_bc_ord = max_deriv_orders   # We assume, that the maximum order of the equation here is 2
+    if global_var.grid_cache is None:
+        raise NameError('Grid cache has not been initialized yet.')
+    
+    bconds = []
+    hardcoded_bc_relative_locations = {0 : None, 1 : (0,), 2 : (0, 1)}
+    tensor_shape = global_var.grid_cache.get('0').shape
+
+    def get_boundary_ind(tensor_shape, axis, rel_loc, old_way = False):
+        return tuple(np.meshgrid(*[np.arange(shape) if dim_idx != axis else min(int(rel_loc * shape), shape-1)
+                   for dim_idx, shape in enumerate(tensor_shape)], indexing = 'ij'))
+    
+    for ax_idx, ax_ord in enumerate(required_bc_ord):
+        for loc_fraction in hardcoded_bc_relative_locations[ax_ord]:
+            indexes = get_boundary_ind(tensor_shape, axis = ax_idx, rel_loc = loc_fraction)
+            coords = np.array([global_var.grid_cache.get(str(idx))[indexes] for idx in np.arange(len(tensor_shape))])
+            vals = global_var.tensor_cache.get(main_var_key)[indexes]
+            
+            coords = torch.from_numpy(coords)
+            vals = torch.from_numpy(vals)
+            bconds.append([coords, vals])    
+    return bconds    
+                
 
 class SoEq(Complex_Structure, moeadd.moeadd_solution):
+    # __slots__ = ['tokens_indep', 'tokens_dep', 'equation_number']
     def __init__(self, pool, terms_number, max_factors_in_term, sparcity = None, eq_search_iters = 100):
         self.tokens_indep = TF_Pool(pool.families_meaningful) #[family for family in token_families if family.status['meaningful']]
         self.tokens_dep = TF_Pool(pool.families_supplementary) #[family for family in token_families if not family.status['meaningful']]
         self.equation_number = np.size(self.tokens_indep.families_cardinality())
         
-        if type(sparcity) != None: self.vals = sparcity 
+        if sparcity is not None: self.vals = sparcity 
         self.max_terms_number = terms_number; self.max_factors_in_term = max_factors_in_term
         self.moeadd_set = False; self.eq_search_operator_set = False# ; self.evaluated = False
         self.def_eq_search_iters = eq_search_iters
         
     def use_default_objective_function(self):
-        from epde.eq_mo_objectives import system_discrepancy, system_complexity_by_terms
-        self.set_objective_functions([system_discrepancy, system_complexity_by_terms])
+        from epde.eq_mo_objectives import system_discrepancy, system_complexity_by_terms, system_complexity_by_factors
+        self.set_objective_functions([system_discrepancy, system_complexity_by_factors])
         
     def set_objective_functions(self, obj_funs):
         '''
@@ -646,12 +749,17 @@ class SoEq(Complex_Structure, moeadd.moeadd_solution):
 #            raise ValueError('Number of iterations is not defied both in method parameter or in object attribute')
         assert self.eq_search_operator_set
         
-        if type(eq_search_iters) == type(None): eq_search_iters = self.def_eq_search_iters
-        if type(sparcity) == type(None): 
+        if eq_search_iters is None: eq_search_iters = self.def_eq_search_iters
+        if sparcity is None: 
             sparcity = self.vals
         else:
             self.vals = sparcity
+
         self.population_size = population_size
+        self.eq_search_evolutionary_strategy.modify_block_params(block_label = 'truncation', 
+                                                                 param_label = 'population_size', 
+                                                                 value = population_size)
+        
         self.structure = []; self.eq_search_iters = eq_search_iters
         token_selection = self.tokens_indep
         
@@ -693,22 +801,6 @@ class SoEq(Complex_Structure, moeadd.moeadd_solution):
                         for i in range(population_size)]
         EA_kwargs['separate_vars'] = separate_vars
         strategy.run(initial_population = population, EA_kwargs = EA_kwargs)
-#        for idx, equation in enumerate(population):
-#            equation.select_target_idx(separate_vars, operator)
-#            equation.check_split_correctness()
-#            operator.get_fitness(equation)
-#            if equation.described_variables in separate_vars:
-#                equation.penalize_fitness(coeff = 0.)
-#        
-#        for idx in range(eq_search_iters):
-#            strict_restrictions = False if idx < eq_search_iters - 1 else True
-#            population = self.equation_opt_iteration(population, self.eq_search_evolutionary_operator, 
-#                                        population_size, idx, separate_vars, strict_restrictions)
-#            for equation in population:
-#                if equation.described_variables in separate_vars:
-#                    equation.penalize_fitness(coeff = 0.)            
-#        population = Population_Sort(population)
-#        del population[1:]
         result = strategy.result
         
         return result[0], result[0].evaluate(normalize = False, return_val=True)[0], result[0].evaluate(normalize = True, return_val=True)[0]
@@ -720,33 +812,18 @@ class SoEq(Complex_Structure, moeadd.moeadd_solution):
                 equation.penalize_fitness(coeff = 0.)           
         population = Population_Sort(population)
         population = population[:population_size]
-#        print(iter_index, population[0].fitness_value, population[0].L0_norm)#, population[0].described_variables, separate_vars, population[0].described_variables in separate_vars)
-#        print('Cache size', np.round(global_var.tensor_cache.consumed_memory / 1024**2, decimals=3), 'MB')
         gc.collect()
-#        if iter_index == 2:
-#            memory_assesment()
-#        time.sleep(10)
-#        prev_population = copy.deepcopy(population) # copy.deepcopy?
         population = evol_operator.apply(population, separate_vars)
         return population
-    
-#    def __eq__(self, other):
-#        assert self.moeadd_set, 'The structure of the equation is not defined, therefore no moeadd operations can be called'
-#        epsilon = 1e-6
-#        if isinstance(other, type(self)):
-#            return np.all(np.abs(self.vals - other.vals) < epsilon)
-#        else:
-#            return NotImplemented
         
     def evaluate(self, normalize = True):
-#            self.evaluated = True
         if len(self.structure) == 1:
             value = self.structure[0].evaluate(normalize = normalize, return_val = True)[0]
         else:
             value = np.sum([equation.evaluate(normalize, return_val = True)[0] for equation in self.structure])
         value = np.sum(np.abs(value))
         return value
-        
+
     @property
     def obj_fun(self):
 #        print('objective functions:', self.obj_funs)
@@ -789,5 +866,26 @@ class SoEq(Complex_Structure, moeadd.moeadd_solution):
             
     def __hash__(self):
         return hash(tuple(self.vals))
-                
         
+    def __deepcopy__(self, memo = None):
+        clss = self.__class__
+        new_struct = clss.__new__(clss)
+        memo[id(self)] = new_struct
+        
+        new_struct.__dict__.update(self.__dict__)
+        
+        for k in self.__slots__:
+                try:
+                    if not isinstance(k, list):
+                        setattr(new_struct, k, copy.deepcopy(getattr(self, k), memo))
+                    else:
+                        temp = []
+                        for elem in getattr(self, k):
+                            temp.append(copy.deepcopy(elem, memo))
+                        setattr(new_struct, k, temp)                            
+                except AttributeError:
+                    pass     
+
+        for idx, eq in enumerate(self.structure):
+            eq.copy_properties_to(new_struct)
+        return new_struct
\ No newline at end of file
